"""
Author: Filip Lux (2023), Masaryk University
Licensed under MIT License
"""

from grpc import insecure_channel, RpcError
from copy import copy
import os
from tqdm import tqdm
from skimage import io, measure
import numpy as np
import math
from pathlib import Path


from . import buckets_with_graphics_pb2 as buckets_with_graphics_pb2
from . import buckets_with_graphics_pb2_grpc as  buckets_with_graphics_pb2_grpc


# GLOBAL PARAMETERS
COLOR = {'red':    0xF54731,      # color of objects
         'green':  0xB4FA41,
         'blue':   0x47CDFF,
         'gray':   0xEEEEEE,
         'yellow': 0xF5BC00,
         'pink':   0xD6559E,
         'black':  0x000000}

CONFIG = {
    'x_scale' : .1,           # scale of the axis x
    'x_shift' : 500,          # shift of the axis x
    'y_scale' : .1,           # scale of the axis y
    'y_shift' : 500,          # shift of the axis y
    'z_scale' : .2,           # scale of the axis z
    'z_shift' : 1000          # shift of the axis z
    }

class TrackingVisualizer:
    
    '''
    tool to send objects from python to Blender
    developed to visualize cell tracking graphs in 3D 
    '''
    
    def __init__(self,
                data_path,
                res_path,
                time,
                idx_min=0,
                idx_max=10000):
        
    
        self.data_path = data_path
        self.res_path = res_path
        self.time = time
        self.idx_min = idx_min
        self.idx_max = idx_max
        
        # TODO: load from config
        self.sequence = os.path.basename(data_path)[:2]
        
        
    def display_gt(self,
                   time=None):
        
        data_dir = os.path.dirname(self.data_path)
        
        # test if gt exists
        gt_path = os.path.join(data_dir,
                               self.sequence + '_GT')
        if not os.path.isdir(gt_path):
            print(f'The sequence do not have a GT files')
            print(gt_path)
            return
        
        time = time if time is not None else self.time

        # send gt to blender
        send_gt(data_dir,
                self.sequence,
                time=self.time,
                first_frame=self.idx_min,
                last_frame=self.idx_max)
        
    def display_candidate_graph(self,
                                time=None):
        
        # test if tracking.txt exists
        tracking_path = os.path.join(self.res_path,
                                     'tracking.txt')
        if not os.path.isfile(tracking_path):
            print(f'The tracking file {tracking_path} do not exists')
            return
        
        time = time if time is not None else self.time

        # send candidate graph to blender
        show_libct(tracking_path,
                   time=time,
                   min_idx=self.idx_min,
                   max_idx=self.idx_max)
    
    def display_solution(self,
                         time=None):
        '''
        displays tracks choosen by a tracker
        '''
        
        # test if tracking.sol exists
        sol_path = os.path.join(self.res_path,
                                     'tracking.sol')
        if not os.path.isfile(sol_path):
            print(f'The solution file {sol_path} do not exists')
            return
        
        time = time if time is not None else self.time
        show_libct_solution(sol_path,
                            time=time,
                            min_idx=self.idx_min,
                            max_idx=self.idx_max)
    
    def display_mistakes(self,
                         time=None):
        '''
        displays errors after ctc evaluation 
        '''
        
                # test if tracking.sol exists
        tra_path = os.path.join(self.res_path,
                                     'TRA_log.txt')
        if not os.path.isfile(tra_path):
            print(f'The tracking log file {tra_path} do not exists')
            return
        
        time = time if time is not None else self.time
        send_res(self.data_path,
                 self.res_path,
                 time=time,
                 first_frame=self.idx_min,
                 last_frame=self.idx_max)
        

def choose_color_and_size(prob):

    '''
    (-inf, -1) green
    (-1, 0) green
    (0, 3) pink
    (3, inf) red

    '''
    
    size = ((prob * 10) // 1) * .1

    size = -prob / 5
    size = max(size, .05)

    # four colors
    if prob <= -2:
        return COLOR['green'], size
    elif prob <= -1:
        return COLOR['yellow'],  size
    elif prob <= 0:
        return COLOR['pink'], size
    else:
        return COLOR['red'], size

def transform_coo(x, y, z):
    '''
    transforms image coordinates to blender coordinates
    '''
    x = (x - CONFIG['x_shift']) * CONFIG['x_scale']
    y = (y - CONFIG['y_shift']) * CONFIG['y_scale']
    z = (z - CONFIG['z_shift']) * CONFIG['z_scale']
    
    return x, y, z


def label2rgb(label):
    '''
    maps an integer label to rgb color
    '''
    
    np.random.seed(label)
    color = list(np.random.choice(range(256), size=3))
    color_rgb = color[2] * 256 * 256 + color[1] * 256 + color[0]
    
    return color_rgb


def get_linear_color(label):
    
    # clip to (0, 1)
    #label = np.minimum(np.maximum(label, 0), 1)

    color = [210, int(255*label), 0]
    color_rgb = color[0] + color[1] * 256 + color[2] * 256 * 256
        
    return color_rgb


def get_label(string):
    
    string = string.replace(']', '').replace('[', '')
    
    time, label = string.strip().split(' ')
    time = time.replace('T=', '')
    label = label.replace('Label=', '').replace('GT_label=', '')
    
    return int(time), int(label)
    


def read_det_log(txt_path):
        
    
    with open(txt_path, 'r') as f:

        lines = f.readlines()[:-2]
        sep = []
        
        for idx, line in enumerate(lines):
            if line[0] == '-':
                sep.append(idx)
        print(sep)
        
        secs = np.split(lines, sep)
        res = [sec[1:] for sec in secs[1:]]
        
        print(len(res))
        
        lines_split, lines_fn, lines_fp = res
        
        
        '----------Splitting Operations (Penalty=5)----------'
        splitting_op = {}
        for line in lines_split:
            
            time, label = get_label(line)
            
            splitting_op[time] = splitting_op.get(time, [])
            splitting_op[time].append(label)

        
        'GT - list'
        '----------False Negative Vertices (Penalty=10)----------'
        fn_vertices = []
        for line in lines_fn:
            
            time, label = get_label(line)
            fn_vertices.append((time, label))
            
        
        '----------False Positive Vertices (Penalty=1)----------'
        fp_vertices = {}
        for line in lines_fp:
            
            time, label = get_label(line)
            
            fp_vertices[time] = fp_vertices.get(time, [])
            fp_vertices[time].append(label)

            
        
        res = splitting_op, fn_vertices, fp_vertices
        
        return res
    
    
def read_tra_log(txt_path):
        
    # for each time frame list index of daughters (time t) and their mothers (time t-1)
    
    keys = set()
    t0_init = set()
    
    tags = []
    with open(txt_path, 'r') as f:

        lines = f.readlines()
        sep = []
        
        for idx, line in enumerate(lines):
            if line[0] == '-':
                sep.append(idx)
        print(sep)
        
        secs = np.split(lines, sep)
        res = [sec[1:] for sec in secs[1:]]
        
        print(len(res))
        
        lines_split, lines_fn, lines_fp, lines_remove, lines_add, lines_sem = res
        
        
        '----------Splitting Operations (Penalty=5)----------'
        splitting_op = {}
        for line in lines_split:
            
            time, label = get_label(line)
            
            splitting_op[time] = splitting_op.get(time, [])
            splitting_op[time].append(label)
            
            keys.add(time)
        
        'GT - list'
        '----------False Negative Vertices (Penalty=10)----------'
        fn_vertices = []
        for line in lines_fn:
            
            time, label = get_label(line)
            fn_vertices.append((time, label))
            
        
        '----------False Positive Vertices (Penalty=1)----------'
        fp_vertices = {}
        for line in lines_fp:
            
            time, label = get_label(line)
            
            fp_vertices[time] = fp_vertices.get(time, [])
            fp_vertices[time].append(label)
            
            keys.add(time)
            
        
        '----------Redundant Edges To Be Deleted (Penalty=1)----------'
        remove_edges = {}
        for line in lines_remove:
            
            start, end = line.strip().split(' -> ')
            time0, label0 = get_label(start)
            time1, label1 = get_label(end)
            
            if not time0 + 1 == time1:
                print(f'Redundant edge: {line}')
                continue
            
            remove_edges[time1] = remove_edges.get(time1, [])
            remove_edges[time1].append((label0, label1))
            
            keys.add(time0)
            keys.add(time1)
            t0_init.add(time0)
            
        'GT - list'
        '----------Edges To Be Added (Penalty=1.5)----------'
        add_edges = []
        skip_edges = []
        for line in lines_add:
            
            start, end = line.strip().split(' -> ')
            time0, label0 = get_label(start)
            time1, label1 = get_label(end)
            
            #assert time0 + 1 == time1, f'{time0} {time1}, {line}'
            if not time0 + 1 == time1:
                skip_edges.append((time0, label0, time1, label1))
                t0_init.add(time0)
            else:
                add_edges.append((time0, label0, time1, label1))
            
            
            
            
        '----------Edges with Wrong Semantics (Penalty=1)----------'
        semantic_edges = {}
        for line in lines_sem:
            
            if line[0] == '=':
                break
            
            start, end = line.strip().split(' -> ')
            time0, label0 = get_label(start)
            time1, label1 = get_label(end)
            
            if not time0 + 1 == time1:
                print(f'Wrong semantics: {line}')
                continue
                
            semantic_edges[time1] = semantic_edges.get(time1, [])
            semantic_edges[time1].append((label0, label1))
            
            keys.add(time0)
            keys.add(time1)
            t0_init.add(time0)
            
        
        print(lines_sem[-1])
        
        res = {}
        res['splitting_op'] = splitting_op
        res['fn_vertices'] = fn_vertices
        res['fp_vertices'] = fp_vertices
        res['remove_edges'] = remove_edges
        res['add_edges'] = add_edges
        res['semantic_edges'] = semantic_edges
        res['skip_edges'] = skip_edges
        
        # depreciated
        # return as a dictionary
        #res = (splitting_op, fn_vertices, fp_vertices, remove_edges, add_edges, semantic_edges), 
        
        return res, (keys, t0_init)
    
    


class BlenderViewer():
    
    def __init__(self,
                 client_name,
                 clientURL="localhost:9085",
                 serverURL="localhost:9083"):
        
        self.clientName = client_name
        self.clientURL = clientURL
        self.serverURL = serverURL
        
        self.comm = buckets_with_graphics_pb2_grpc.ClientToServerStub( insecure_channel(serverURL) )
        
        self.sphere_buckets = {}
        self.line_buckets = {}
        
        # self.__greetings()
        
        
    def __greetings(self):
        clientGreeting = buckets_with_graphics_pb2.ClientHello()
        clientGreeting.clientID.clientName = self.clientName
        clientGreeting.returnURL = self.clientURL
        self.comm.introduceClient(clientGreeting)
        
    def put_sphere(self, coo0, object_id=128, color=None, size=0.4):
        x0, y0, z0, t0 = coo0
        
        if color is None:
            color = label2rgb(object_id)
        
        sphParams = buckets_with_graphics_pb2.SphereParameters()
        sphParams.centre.x = x0
        sphParams.centre.y = y0
        sphParams.centre.z = z0
        sphParams.radius = size
        sphParams.colorXRGB = color
        
        # create bucket if not exists
        if object_id not in self.sphere_buckets.keys():
            
            bucket = buckets_with_graphics_pb2.BucketOfSpheres()
            bucket.clientID.clientName = self.clientName
            bucket.bucketID = object_id

            bucket.time = t0
            
            self.sphere_buckets[object_id] = bucket
            
        # start bucket
        self.sphere_buckets[object_id].spheres.append(sphParams)
                  
        #self.sphere_buckets[object_id].spheres.append(sphParams)

        
    def put_vector(self, coo0, coo1, object_id=128, color=None, radius=1):
        
        # coordinates
        x0, y0, z0, t0 = coo0
        x1, y1, z1, t1 = coo1
        
        if color is None:
            color = label2rgb(object_id)
                
        #
        line = buckets_with_graphics_pb2.LineParameters()
        line.startPos.x = x0
        line.startPos.y = y0
        line.startPos.z = z0
        line.endPos.x = x1
        line.endPos.y = y1
        line.endPos.z = z1
        line.radius = radius
        line.colorXRGB = color
        
        # create bucket if not exists
        if object_id not in self.line_buckets.keys():
        
        
            bucket = buckets_with_graphics_pb2.BucketOfLines()
            bucket.clientID.clientName = self.clientName
            bucket.bucketID = object_id

            bucket.time = t0
            
            self.line_buckets[object_id] = bucket
            
        self.line_buckets[object_id].lines.append(line)

        
    
    def send_buckets(self):
        
        self.__greetings()
        
        # send spheres
        sphere_list = []
        for key in self.sphere_buckets.keys():
            sphere_list.append(self.sphere_buckets[key])
            
        self.comm.addSpheres(iter(sphere_list))
        self.sphere_buckets = {}
        
        
        # send lines
        line_list = []
        for key in self.line_buckets.keys():
            line_list.append(self.line_buckets[key])
            
        self.comm.addLines(iter(line_list))
        self.line_buckets = {}
        

def read_man_gt(path,
                seq,
                first_frame=0,
                last_frame=10000):
    
        
    # gt path
    gt_path = os.path.join(path, f'{seq}_GT', 'TRA')
    assert os.path.isdir(gt_path), gt_path
    
     # list of available files
    tif_files = [os.path.join(gt_path, file) for file in os.listdir(gt_path) if '.tif' in file]
    tif_files.sort()

    gt_coordinates = {}
    for frame_idx, file_path in enumerate(tqdm(tif_files)):
        if not (first_frame < frame_idx < last_frame):
            continue
        
        
        gt = io.imread(file_path)
        
        frame_dict = {}
        
        regions = measure.regionprops(gt)
        
        for reg in regions:
            frame_dict[reg.label] = reg.centroid
            
        gt_coordinates[frame_idx] = frame_dict
        
    return gt_coordinates


def send_det(path, suffix, seq, time, first_frame=0, last_frame=10000, dz=1, gt_dict=None):
    
    
    name = os.path.basename(path)
    
    new_label = 3000
    blender = BlenderViewer(f'{name}_{seq}_{suffix}')

    
    # tracking path
    res_path = os.path.join(f'{path}_{suffix}', f'{seq}_RES')
    assert os.path.isdir(res_path), res_path
    
    # gt path
    gt_path = os.path.join(f'{path}', f'{seq}_GT', 'TRA')
    assert os.path.isdir(gt_path), gt_path


    # list of available files
    tif_files = [os.path.join(res_path, file) for file in os.listdir(res_path) if '.tif' in file]
    tif_files.sort()
    log_file = os.path.join(res_path, 'DET_log.txt')
    
    
    
    if not os.path.isfile(log_file):
        print(f'tra res file {log_file} do not exists. Run DETMeasure first')
        return
    
    res = read_det_log(log_file)
   
    # result of DETMeasure
    splitting_op, fn_vertices, fp_vertices = res
    
    
    if gt_dict is None:
        gt_dict = read_man_gt(path, seq, first_frame, last_frame)
        
    detections = {}
    
    # TODO: iterate only over the <keys>
    for frame_idx in tqdm(range(first_frame, np.minimum(len(tif_files), last_frame))):

        
        # read gt and res
        gt = io.imread(tif_files[frame_idx])
        if gt.dtype != int:
            gt = gt.astype(int)
            #print(tif_files[frame_idx])
        
        regions = measure.regionprops(gt)
        
        # get regions
        regions_lab = {}
        for reg in regions:
            regions_lab[int(reg.label)] = reg
            
        
        # split
        for label in splitting_op.get(frame_idx, []):
            
            reg = regions_lab[label]
            
            x1, y1 = reg.centroid
            x1, y1, t1 = transform_coo(x1, y1, frame_idx)
            
            blender.put_sphere((x1, y1, t1, time),
                   object_id=1,
                   color=COLOR['red'])
            
        # fp_vertices
        for label in fp_vertices.get(frame_idx, []):
                        
            reg = regions_lab[label]
            
            x1, y1 = reg.centroid
            x1, y1, t1 = transform_coo(x1, y1, frame_idx)
            
            blender.put_sphere((x1, y1, t1, time),
                   object_id=2,
                   color=COLOR['yellow'],
                   size=0.2)
        
    #fn_vertices
    for frame_idx, idx in fn_vertices:

        gt_name = os.path.join(gt_path, f'man_track{frame_idx:04}.tif')
        assert os.path.isfile(gt_name), gt_name

        gt = io.imread(gt_name)

        x1, y1 = gt_dict[frame_idx][idx]
        x1, y1, t1 = transform_coo(x1, y1, frame_idx)

        #print('sphere', time1, label1, x1, y1, t1)
        blender.put_sphere((x1, y1, t1, time),
               object_id=5,
               color=COLOR['blue'], size=.2)

       
    

    
    blender.send_buckets()
    
    
    
def init_tracks(tif_files, first_frame, dz=1):
    
    tracks = {}
    
    frame_idx = first_frame-1
    
    if frame_idx >=0 :
    
        gt = io.imread(tif_files[frame_idx])
        regions = measure.regionprops(gt)

        for reg in regions:

            label = reg.label
            x, y = reg.centroid
            x, y, t = transform_coo(x, y, frame_idx)

            tracks[label] = x, y, t
        
    return tracks
        

def send_gt(path,
            seq,
            time,
            first_frame=0,
            last_frame=10000,
            individual_tracks=False,
            radius=.5):

    """
    Displays GT in a Blender


    Parameters
    ----------
    path : str
        path to the solution
    seq : str
        sequence, e.g. '01', '02', ..
    time : int
        Blender timepoint 
    first_frame : int 
        first frame index that is send to the Blender
    last_frame : int 
        last frame index that is send to the Blender
    individual_tracks : bool
        True when asign an unique ID to each track
    radius : float
        radius of the spheres

    Returns
    -------

    None

    """

    first_frame = first_frame + 1
    
    
    new_label = 3000
    blender = BlenderViewer(f'BF-C2DL-HSC_gt')

    
    # tracking path
    img_path = os.path.join(path, seq)
    assert os.path.isdir(img_path), img_path
    
    # gt path
    gt_path = os.path.join(path, f'{seq}_GT', 'TRA')
    assert os.path.isdir(gt_path), gt_path
    

    # list of available files
    tif_files = [os.path.join(gt_path, file) for file in os.listdir(gt_path) if '.tif' in file]
    tif_files.sort()
    txt_file = os.path.join(gt_path, 'man_track.txt')
    assert os.path.isfile(txt_file), txt_file
    
    # analize mother and daughter cells
    daughters = get_daughters_dict(txt_file)
    
    # fixed label
    object_id = 1
    

    tracks = init_tracks(tif_files, first_frame, dz=CONFIG['z_scale'])
    
    for frame_idx in tqdm(range(first_frame, np.minimum(len(tif_files), last_frame))):
        
        # read gt and res
        gt = io.imread(tif_files[frame_idx])
        regions = measure.regionprops(gt)
        
        for reg in regions:
            
            x1, y1 = reg.centroid
            x1, y1, t1 = transform_coo(x1, y1, frame_idx)
            
            label = reg.label
            
            if individual_tracks:
                object_id = label
            
            
            if label in tracks.keys():
                
                # find previous stage
                x0, y0, t0 = tracks[label]
                
                # put gray line
                blender.put_vector((x0, y0, t0, time),
                                   (x1, y1, t1, time),
                                   object_id=object_id,
                                   color=COLOR['gray'],
                                   radius=radius)
                
                # put sphere
                blender.put_sphere((x0, y0, t0, time),
                                   object_id=2,
                                   color=COLOR['gray'],
                                   size=.15)
                
            
            
            else:
                assert label in daughters[frame_idx].keys(), f'{label} {daughters[frame_idx].keys()}'
                
                m_idx = daughters[frame_idx][label]
                x0, y0, t0 = tracks[m_idx]
                
                # put gray line
                blender.put_vector((x0, y0, t0, time),
                                   (x1, y1, t1, time),
                                   object_id=object_id,
                                   color=COLOR['gray'],
                                   radius=radius)
                
                # put sphere
                blender.put_sphere((x0, y0, t0, time),
                                   object_id=object_id,
                                   color=COLOR['gray'],
                                   size=.15)
                
            # update tracks    
            tracks[label] = x1, y1, t1
                
            
    blender.send_buckets()
    
    



def send_res(data_path,
             res_path,
             time,
             first_frame=0,
             last_frame=10000,
             gt_dict=None):

    """
    Displays RES file in a Blender


    Parameters
    ----------
    path : str
        path to the result
    time : int
        timepoint 
    first_frame : int 
        first frame index that is send to the Blender
    last_frame : int 
        last frame index that is send to the Blender
    gt_dict : None, dict
        depreciated
        dictionary describing the res_tra.txt file

    Returns
    -------

    None

    """
    
    name = os.path.basename(res_path)
    
    new_label = 3000
    blender = BlenderViewer(f'{res_path}')

    # tracking path
    assert os.path.isdir(res_path), res_path

    # list of available files
    tif_files = [os.path.join(res_path, file) for file in os.listdir(res_path) if '.tif' in file]
    tif_files.sort()
    tra_path = os.path.join(res_path, 'TRA_log.txt')
    
    if not os.path.isfile(tra_path):
        print(f'TRA res file {tra_path} do not exists. Run TRAMeasure procedure first.')

        # TODO: run TRA measure rather than stop the procedure
        return
    
    res, (keys, t0_init) = read_tra_log(tra_path)
   
    # result of TRAMeasure
    splitting_op = res['splitting_op']
    fn_vertices = res['fn_vertices']
    fp_vertices = res['fp_vertices']
    remove_edges = res['remove_edges']
    add_edges = res['add_edges']
    semantic_edges = res['semantic_edges']
    skip_edges = res['skip_edges']
    
    if gt_dict is None:
        path = os.path.dirname(data_path)
        seq = os.path.basename(data_path)[:2]
        gt_dict = read_man_gt(path, seq)
        
    detections = {}
    
    # initialize detections
    if first_frame > 0:
        frame_idx = first_frame-1
        gt = io.imread(tif_files[frame_idx])
        regions = measure.regionprops(gt)
        for reg in regions:

            x1, y1 = reg.centroid
            x1, y1, t1 = transform_coo(x1, y1, frame_idx)

            detections[reg.label] = x1, y1, t1
    
    # TODO: iterate only over the <keys>
    for frame_idx in tqdm(range(first_frame, np.minimum(len(tif_files), last_frame))):
                
        if frame_idx not in keys:
            continue
        
        # read gt and res
        gt = io.imread(tif_files[frame_idx])
        regions = measure.regionprops(gt)
        
        # get regions
        regions_lab = {}
        for reg in regions:
            regions_lab[int(reg.label)] = reg
            
        
        # split
        for label in splitting_op.get(frame_idx, []):
            
            reg = regions_lab[label]
            
            x1, y1 = reg.centroid
            x1, y1, t1 = transform_coo(x1, y1, frame_idx)
            
            blender.put_sphere((x1, y1, t1, time),
                   object_id=1,
                   color=COLOR['red'])
            

            
        # fp_vertices
        for label in fp_vertices.get(frame_idx, []):
            
            reg = regions_lab[label]
            
            x1, y1 = reg.centroid
            x1, y1, t1 = transform_coo(x1, y1, frame_idx)
            
            blender.put_sphere((x1, y1, t1, time),
                   object_id=2,
                   color=COLOR['yellow'],
                   size=0.2)
            
        # remove_edges
        for label0, label1 in remove_edges.get(frame_idx, []):
            
            reg = regions_lab[label1]
            
            x0, y0, t0 = detections[label0]
            
            x1, y1 = reg.centroid
            x1, y1, t1 = transform_coo(x1, y1, frame_idx)
            
            blender.put_vector((x0, y0, t0, time),
                               (x1, y1, t1, time),
                               object_id=3,
                               color=COLOR['red'])
            

        
            

            
        # semantic_edges
        for label0, label1 in semantic_edges.get(frame_idx, []):
            
            if label0 != label1:
                # extra split
                object_id = 4
                color = COLOR['pink']
            else:
                # missing split
                object_id = 7
                color=COLOR['green']
            
            reg = regions_lab[label1]
            
            x0, y0, t0 = detections[label0]
            
            x1, y1 = reg.centroid
            x1, y1, t1 = transform_coo(x1, y1, frame_idx)
            
            blender.put_vector((x0, y0, t0, time),
                               (x1, y1, t1, time),
                               object_id=object_id,
                               color=color)
            
            blender.put_sphere((x0, y0, t0, time),
                   object_id=object_id,
                   color=color,
                   size=0.2)
            
        # initialize previous detections    
        if frame_idx in t0_init:
            for reg in regions:
                
                x1, y1 = reg.centroid
                x1, y1, t1 = transform_coo(x1, y1, frame_idx)
                
                detections[reg.label] = x1, y1, t1
                
    # missing objects
    
    # skip_edges
    print('the number of skip edges', len(skip_edges))
    for time0, label0, time1, label1 in tqdm(skip_edges):
        if not (first_frame <= time0) and (time2 < last_frame ):
            continue

        x0, y0 = gt_dict[time0][label0]
        x0, y0, t0 = transform_coo(x0, y0, time0)

        x1, y1 = gt_dict[time1][label1]
        x1, y1, t1 = transform_coo(x1, y1, time1)

        blender.put_vector((x0, y0, t0, time),
                           (x1, y1, t1, time),
                           object_id=6,
                           color=COLOR['green'])

    # fn_vertices
    print('the size of FN_verticies', len(fn_vertices))
    for time1, label1 in tqdm(fn_vertices):
        if not (first_frame <= time1 < last_frame):
            continue

        x1, y1 = gt_dict[time1][label1]        
        x1, y1, t1 = transform_coo(x1, y1, time1)

        blender.put_sphere((x1, y1, t1, time),
               object_id=5,
               color=COLOR['blue'], size=.2)
        
    # add_edges
    print('the number of mising edges', len(add_edges))
    for time0, label0, time1, label1 in tqdm(add_edges):
        if not (first_frame <= time1 < last_frame - 1):
            continue

        x0, y0 = gt_dict[time0][label0]
        x0, y0, t0 = transform_coo(x0, y0, time0)
        
        x1, y1 = gt_dict[time1][label1]
        x1, y1, t1 = transform_coo(x1, y1, time1)

        blender.put_vector((x0, y0, t0, time),
                           (x1, y1, t1, time),
                           object_id=6,
                           color=COLOR['blue'])
        
        #print('vector', time1, label1, x1, y1, t1)
       
    

    
    blender.send_buckets()
    
    
def get_daughters_dict(txt_file):
    '''
    Returns
    -------

    dict:
    daughters = {frame_idx:
                        { daughter_index : mother_index}
                }

    '''

    
    daughters = {}
    with open(txt_file, 'r') as f:

        for line in f.readlines():
            d_idx, f_idx, _, m_idx = line.strip().split(' ')
            
            f_idx = int(f_idx)
            d_idx = int(d_idx)
            m_idx = int(m_idx)
            
            daughters[f_idx] = daughters.get(f_idx, {})
            daughters[f_idx][d_idx] = m_idx
            
    return daughters


def show_libct(file_path,
               z_shift=0,
               description='',
               time=10,
               object_id=5,
               min_idx=0,
               max_idx=2000):
    '''
    Displays candidate graph 

    Parameters
    ----------

    file_path : str
        path to tracking.txt file, in a "libct" format
    z_shift : int
        change of the position in a z axis due to the consitency
        with a ctlib solver that expects the tracking at a zero level
    description : str
        string indentificator of the bledner objects
    time : int
        timepoint in a Blender to display objects
    object_id : int
        Not used
    min_idx : int                  NOT IMPLEMENTED
        index of starting frame  
    max_idx : int                  NOT IMPLEMENTED
        index of the first missing frame
        (displays frames {min_idx, min_idx + 1, ..., max_idx - 1 })

    '''
    
    
    assert os.path.isfile(file_path), file_path
    if min_idx >= max_idx:
        print('Nothing to display')
        return
    
    blender = BlenderViewer(f'{file_path}_{description}')
    
    e_id = 5
    e_count = 1
    
    vertices = {}
    
    print(f'range {min_idx} <= frame_idx <= {max_idx}')
    
    with open(file_path, 'r') as f:
        
        # show vertices
        for line in tqdm(f.readlines()):
            key = line.split(' ')[0]
            if key == 'H':
                
                _, frame_idx, unique_id, cost, x, y = line.split(' ')
                
                # test range
                if not (min_idx <= int(frame_idx) < max_idx):
                    continue
                    
                # transform coordinates
                x, y, z = transform_coo(float(x), float(y), int(frame_idx) + z_shift)
                # save vertex
                vertices[unique_id] = (x, y, z)
                
                
                
                color, size = choose_color_and_size(float(cost))
                blender.put_sphere((x, y, z, time),
                   object_id=1,
                   color=color,
                   size=size * .1)
                

                
            elif key == 'APP':
                
                _, event_id, vertex_id, cost = line.split(' ')
                if vertex_id not in vertices.keys():
                    continue
                
                x1, y1, frame_idx = vertices[vertex_id]
                x2, y2 = x1 + 20, y1 + 20
                
                # filter out forbidden APP events
                if float(cost) > 1:
                    continue
                
                color, size = choose_color_and_size(float(cost))
                blender.put_vector(
                           (x1, y1, frame_idx, time),
                           (x2, y2, frame_idx, time),
                           object_id=2,
                           color=color,
                           radius=size)
                
                
            elif key == 'DISAPP':
                _, event_id, vertex_id, cost = line.split(' ')
                if vertex_id not in vertices.keys():
                    continue
                    
                x1, y1, z = vertices[vertex_id]
                x2, y2 = x1 + 20, y1 - 20
                
                # filter out forbidden APP events
                if float(cost) > 1:
                    continue
                
                color, size = choose_color_and_size(float(cost))
                blender.put_vector(
                           (x1, y1, z, time),
                           (x2, y2, z, time),
                           object_id=3,
                           color=color,
                           radius=size)
                                
            elif key == 'MOVE':
                _, event_id, left_id, right_id, cost = line.split(' ')
                if any([v_id not in vertices.keys() for v_id in [left_id, right_id]]):
                    continue
                
                x1, y1, z1 = vertices[left_id]
                x2, y2, z2 = vertices[right_id]
                
                
                color, size = choose_color_and_size(float(cost))
                blender.put_vector(
                           (x1, y1, z1, time),
                           (x2, y2, z2, time),
                           object_id=5,
                           color=color,
                           radius=size)
                
            elif key == 'DIV':
                _, event_id, mother_id, left_id, right_id, cost = line.split(' ')
                if any([v_id not in vertices.keys() for v_id in [mother_id, left_id, right_id]]):
                    continue
                    
                x1, y1, frame_idx1 = vertices[left_id]
                x2, y2, frame_idx2 = vertices[right_id]
                
                prob = math.exp(-float(cost))
                color, size = choose_color_and_size(float(cost))
                blender.put_vector(
                           (x1, y1, frame_idx1, time),
                           (x2, y2, frame_idx2, time),
                           object_id=4,
                           color=0x1122FF,
                           radius=size)
                
                
            else:
                assert False, line
                
                
            # limit number of vectors in a bucket
            e_count += 1
            if e_count % 50000 == 0:
                e_id += 1
                print('sending buckets')
                blender.send_buckets()
                #blender = BlenderViewer(f'{file_path}_{description}_{e_id}')
                
                
    blender.send_buckets()
                
    return


def show_libct_solution(sol_path,
               z_shift=0,
               time=10,
               object_id=5,
               min_idx=0,
               max_idx=2000):
    '''
    Displays candidate graph 

    Parameters
    ----------

    file_path : str
        path to tracking.txt file, in a "libct" format
    z_shift : int
        change of the position in a z axis due to the consitency
        with a ctlib solver that expects the tracking at a zero level
    time : int
        timepoint in a Blender to display objects
    object_id : int
        Not used

    '''
    
    tra_path = str(sol_path).replace('tracking.sol', 'tracking.txt')
    
    assert os.path.isfile(tra_path), tra_path
    assert os.path.isfile(sol_path), sol_path
    
    # COLORS 
    
    # get vertex map from 'tracking.txt'
    vertex_map = {} # vertex_id : (x, y, frame_idx)
    with open(tra_path, 'r') as f:
        
        # show vertices
        for line in tqdm(f.readlines()):
            key = line.split(' ')[0]
            if key == 'H':
                
                _, frame_idx, unique_id, cost, x, y = line.split(' ')
                
                # test range
                if not (min_idx <= int(frame_idx) < max_idx):
                    continue
                
                x, y, z = transform_coo(float(x), float(y), int(frame_idx) + z_shift)
                vertex_map[unique_id] = (x, y, z)

    
    blender = BlenderViewer(f'solution_{sol_path}')
    e_count = 1
    
    with open(sol_path, 'r') as f:
        
        # show vertices
        for line in tqdm(f.readlines()):
            
            line = line.rstrip()
            
            key, unique_id = line.split(' ')
            if key == 'H':
                if unique_id not in vertex_map.keys():
                    continue
                
                x, y, z = vertex_map[unique_id]
    
                blender.put_sphere((x, y, z, time),
                   object_id=1,
                   color=COLOR['black'], size=.05)
                
            elif key == 'APP':
                
                vertex_id = '1' + str(unique_id[1:])
                if vertex_id not in vertex_map.keys():
                    continue
                    
                x, y, z = vertex_map[vertex_id]
    
                blender.put_sphere((x, y, z, time),
                   object_id=2,
                   color=COLOR['yellow'], size=.2)
                

            elif key == 'DISAPP':

                vertex_id = '1' + str(unique_id[1:])
                if vertex_id not in vertex_map.keys():
                    continue
                    
                x, y, z = vertex_map[vertex_id]
    
                blender.put_sphere((x, y, z, time),
                   object_id=3,
                   color=COLOR['red'], size=.2)
        
            elif key == 'MOVE':
                
                vertex1_id = str(unique_id)[1:12]
                vertex2_id = str(unique_id)[12:]
                if any([v_id not in vertex_map.keys() for v_id in [vertex1_id, vertex2_id]]):
                    continue
                
                
                assert len(vertex1_id) == 11, vertex1_id
                assert len(vertex2_id) == 11, vertex2_id
                
                x1, y1, z1 = vertex_map[vertex1_id]
                x2, y2, z2 = vertex_map[vertex2_id]
                
                blender.put_vector(
                           (x1, y1, z1, time),
                           (x2, y2, z2, time),
                           object_id=4,
                           color=COLOR['black'],
                           radius=.3)
                
            elif key == 'DIV':
                
                mother_id = str(unique_id[1:12])
                daughter1_id = str(unique_id[12:23])
                daughter2_id = str(unique_id[23:34])
                if any([v_id not in vertex_map.keys() for v_id in [mother_id, daughter1_id, daughter2_id]]):
                    continue

                x, y, frame_idx = vertex_map[mother_id]
    
                blender.put_sphere((x, y, frame_idx, time),
                   object_id=5,
                   color=COLOR['blue'], size=.2)

                # show lines to daughters


                for d_id in [daughter1_id, daughter2_id]:

                    xd, yd, frame_idx_d = vertex_map[d_id]

                    blender.put_vector(
                       (x, y, frame_idx, time),
                       (xd, yd, frame_idx_d, time),
                       object_id=5,
                       color=COLOR['blue'],
                       radius=.3)
                
                
            else:
                assert False, line
                
                
            # limit number of vectors in a bucket
            e_count += 1
            if e_count % 50000 == 0:
                print('sending buckets')
                blender.send_buckets()
                #blender = BlenderViewer(f'solution_{sol_path}')
                
                e_count = 1
                                
                
    blender.send_buckets()
                
    return
    