"""
Author: Filip Lux (2023), Masaryk University
Licensed under MIT License
"""

from pathlib import Path
import numpy as np
import pandas as pd
from tqdm import tqdm

def compute_divergence(path, plot=False):
    
    '''
    offset_path = path / 'wavefunc_offset.csv'
    seg_path = path / 'wavefunc_seg.csv'
    vertex_prob_path =  path / 'vertex_prob.csv'
    '''
    
    offset_path = path / 'tra_offsets.csv'
    seg_path = path / 'seg_offsets.csv'
    vertex_prob_path =  path / 'vertex_prob.csv'


    assert offset_path.exists(), offset_path
    assert seg_path.exists()
    assert vertex_prob_path.exists(), vertex_prob_path
    

    df_off = pd.read_csv(offset_path)
    df_seg = pd.read_csv(seg_path)
    df_vprob = pd.read_csv(vertex_prob_path)

    samples = []

    times = np.unique(df_off.time)
    times[::-1].sort()

    for time_offset in tqdm(times):

        # correction of the offset time
        time = time_offset - 1

        # get slices of time 'time'
        seg_prev = df_seg[df_seg.time == time - 1]
        off_curr = df_off[df_off.time == time_offset]

        seg_curr = df_seg[df_seg.time == time]

        assert len(off_curr) == len(seg_curr)

        if plot:
            plot_offsets(seg, off)

        samples += compute_kl_multivariate(seg_prev, off_curr, time)


    edge_prob_df = pd.DataFrame(samples, columns=['time_curr', 'time_prev', 'label_curr', 'label_prev', 'divergence'])
    edge_prob_df['cost'] = np.log(edge_prob_df.divergence)

    # store edge and vertex probs as .csv files
    divergence_path = path / "edge_prob_divergence.csv"
    
    print(f'storing {divergence_path}')
    edge_prob_df.to_csv(
        divergence_path,
        index=False,
    )
    
def compute_divergence_v2(path, plot=False):
    
    offset_path = path / 'tra_offsets.csv'
    seg_path = path / 'seg_offsets.csv'
    vertex_prob_path =  path / 'vertex_prob.csv'


    assert offset_path.exists(), offset_path
    assert seg_path.exists()
    assert vertex_prob_path.exists(), vertex_prob_path
    
    divergence_path = path / "edge_prob_divergence.csv"
    
    if divergence_path.exists():
        print(divergence_path, 'already exists')
        return 
    

    df_off = pd.read_csv(offset_path)
    df_seg = pd.read_csv(seg_path)
    df_vprob = pd.read_csv(vertex_prob_path)

    samples = []

    times = np.unique(df_off.time)
    times[::-1].sort()

    for time_curr in tqdm(times):

        # correction of the offset time
        time_prev = time_curr - 1

        # get slices of time 'time'
        seg_prev = df_seg[df_seg.time == time_prev]
        off_curr = df_off[df_off.time == time_curr]

        seg_curr = df_seg[df_seg.time == time_curr]
        
        assert len(off_curr) == len(seg_curr)

        if plot:
            plot_offsets(seg, off)

        samples += compute_kl_multivariate(seg_prev, off_curr, time_curr)


    edge_prob_df = pd.DataFrame(samples, columns=['time_curr', 'time_prev', 'label_curr', 'label_prev', 'divergence'])
    edge_prob_df['cost'] = np.log(edge_prob_df.divergence)

    # store edge and vertex probs as .csv files   
    print(f'storing {divergence_path}')
    edge_prob_df.to_csv(
        divergence_path,
        index=False,
    )
    
    
def kl_divergence(mu1, mu2, sigma1, sigma2):
    return np.log1p(sigma2/sigma1) + (sigma1**2 + (mu1 - mu2) ** 2) / (2 * sigma2 ** 2) - .5


def print_kl(seg, off):
    
    samples = []
    for i, line in enumerate(seg.values):
        
        _, label1, mx1, my1, sx1, sy1 = line
        
        
        for j, line in enumerate(off.values):
            _, label2, mx2, my2, sx2, sy2 = line
            
            kl = kl_divergence(mx1, mx2, sx1, sx2) + kl_divergence(my1, my2, sy1, sy2)
            
            samples.append([label1, label2, kl])
            
    return pd.DataFrame(samples, columns=['label from', 'label to', 'divergence'])


def compute_kl_multivariate(seg, off, time, divergence_thr=50, distance_thr=100):
    
    samples = []
    for i, line in enumerate(seg.values):
        
        # time prev
        _, label_seg_prev, mx, my, sx, sy, c00, c01, c10, c11 = line
        
        S0 = np.array([[c00, c01], [c10, c11]])
        m0 = np.array([mx, my])
        
        for j, line in enumerate(off.values):
            
            # time curr
            _, label_off_curr, mx, my, sx, sy, c00, c01, c10, c11 = line
            
            S1 = np.array([[c00, c01], [c10, c11]])
            m1 = np.array([mx, my])
            
            if euklidian_distance(m0, m1) > distance_thr:
                continue
            
            kl_divergence = kl_mvn(m0, S0, m1, S1)
            
            if kl_divergence < divergence_thr:
                # columns=['time_curr', 'time_prev', 'label_curr', 'label_prev', 'divergence']
                samples.append([time, time-1, int(label_off_curr), int(label_seg_prev), kl_divergence])
            
    #return pd.DataFrame(samples, columns=['time_curr', 'time_prev', 'label_curr', 'label_prev', 'divergence'])
    return samples


def euklidian_distance(m0, m1):
    return np.linalg.norm(m0 - m1)
            
    
def kl_mvn(m0, S0, m1, S1):
    """
    Kullback-Liebler divergence from Gaussian pm,pv to Gaussian qm,qv.
    Also computes KL divergence from a single Gaussian pm,pv to a set
    of Gaussians qm,qv.
    

    From wikipedia
    KL( (m0, S0) || (m1, S1))
         = .5 * ( tr(S1^{-1} S0) + log |S1|/|S0| + 
                  (m1 - m0)^T S1^{-1} (m1 - m0) - N )
    """
    # store inv diag covariance of S1 and diff between means
    N = m0.shape[0]
    iS1 = np.linalg.inv(S1)
    diff = m1 - m0

    # kl is made of three terms
    tr_term   = np.trace(iS1 @ S0)
    det_term  = np.log(np.linalg.det(S1)/np.linalg.det(S0)) #np.sum(np.log(S1)) - np.sum(np.log(S0))
    quad_term = diff.T @ np.linalg.inv(S1) @ diff #np.sum( (diff*diff) * iS1, axis=1)
    #print(tr_term,det_term,quad_term)
    return .5 * (tr_term + det_term + quad_term - N) 